#ifndef BIQUAD__H
#define BIQUAD__H

#define BUTTERWORTH     /* not worth optimizing out */

#include <complex>
typedef std::complex<double> C;

class biquad{
public:
  double c0, c1, c2, d1, d2;
  double ws1, ws2;      // remembered weighted sums

// trivial constructor:
  biquad(){};

// almost-trivial constructor:  all-pass, with specified gain:
  biquad(double gain)
  : c0(gain), c1(0), c2(0), d1(0), d2(0),
  ws1(0), ws2(0)
  {};

// useful constructor, based on explicit coefficients
//  c0, c1, c2 are the coefficients in the numerator
// (1), d1, d2 are the coefficients in the denominator
  biquad(const double _c0, const double _c1, const double _c2,
                           const double _d1, const double _d2)
  : c0(_c0), c1(_c1), c2(_c2), /* d0 does not exist */ d1(_d1), d2(_d2),
    ws1(0), ws2(0)
  {}

  void normalize();
  void dump_coeff();

// the big update-step:
  double operator()(const double Vin);

// transfer function, calculated via z transform:
  C xfunc(const C z);
};

// Notch filter.  Derived class.
// angle is the angle in the z-plane, in radians
// i.e. 2 * pi * notch_freq / sampling_freq
// i.e. 1 * pi * notch_freq / Nyquist_freq
class notch : public biquad{
public:
  notch(const double angle);
};

#ifdef BUTTERWORTH
#include "iir_lp.h"

// Two-pole Butterworth bandpass filter.  Derived class.
// angle specifies the corner
// as an angle in the z-plane, in radians
// i.e. 2 * pi * corner_freq / sampling_freq
// i.e. 1 * pi * corner_freq / Nyquist_freq
class butterworth_2_lp : public biquad{
public:

// constructors:
  butterworth_2_lp();
  butterworth_2_lp(const double angle);
  void set_corner(const double angle);
};


#endif

#endif /* BIQUAD__H */
