//////////////////////////////////////////////////////////////////////
// Evaluate the Gaussian normal distribution

#include "gauss.h"	/* implies cmath */
using namespace std;

const double norm1(1/sqrt(2*M_PI));
const double roothalf(sqrt(.5));

double normDen(const double x, const double mean, const double stdev){
  double foo((x - mean)/stdev);
  return norm1 * exp(-.5*foo*foo) / stdev;
}

double normCum(const double x, const double mean, const double stdev){
  return .5 + .5 * erf(roothalf*(x-mean)/stdev);
}


// This is good for research, because it allows the
// distribution to be shifted by an amount x
// relative to the symmetrical position.
// Not really needed for turbid.
double normHist(const double x, const double mean,
	const double stdev, const int nslot){
  double ss(0);
  int loslot(nslot/2);
  int hislot(nslot - loslot - 1);
  double left(0);
  double ptot(0);
  for (int ii = -loslot; ii <= hislot; ii++) {
    double right(1);
    if (ii != hislot) right = normCum(x+ii+.5, mean, stdev);
    double pr = right - left;
    if (pr) {
      ss += -pr * log2(pr);
      ptot += pr;	// good sanity check
    }
    left = right;	// for next time
  }
  return ss;
}

// Mostly for research.
double normEntTest(const double x, const double mean, const double stdev){
  if (stdev > 3.29) return log2(stdev) + pi2e;
  return normHist(x, mean, stdev, 21);
}

// This is mostly for internal use;
// see operator() for the main user interface.
double entropy_digitized_gaussian::direct(
                double const sigq, int const more){

// Going out 4 sigma is more than we need, but
// too much more.  It is enough to guarantee that
// whatever probability is out in the wings does
// not contribute appreciably to the entropy.
// Note that the summand (aka integrand) has a
// factor of x^2 in it, as in x^2 exp(-x^2),
// so we have to go out a little farther than
// you might have guessed.
  double howfar(4.);            // in units of sigma

  stot = 0;
  ptot = 0;
  int loslot(floor(- howfar * sigq) - more);
  nterms = -2*loslot + 1;

  double left(0);
  for (int ii = loslot; ii <= 0; ii++) {
    double right = normCum(ii+.5, 0, sigq);
    double prob = right - left;

    if (prob) {
      double snow = -prob * log2(prob);
      if (ii) {
        stot += 2. * snow;
        ptot += 2. * prob;	// good sanity check
      } else {
        stot += snow;
        ptot += prob;	// good sanity check
      }
    }
    left = right;	// for next time
  }
  return stot;
}

// This is the main thing you call.
// We guarantee precision within 0.1% of the answer.
// That's for the analytic approximation, with directly==4.
// The direct calculation is even more accurate than that.
// See ./gauss-test.c
double entropy_digitized_gaussian::operator()(
                double const sigq, int const more){
  if (sigq <= directly) return direct(sigq, more);

// Rather than the direct sum, return the
// analytic approximation.
  stot = log2(sigq) + pi2e;
  ptot = 0;
  nterms = 0;
  return stot;
}
