/////////////////////////

#include <iostream>
#include <string>
#include <stdlib.h>     /* for atoi() */
#include <map>
#include <sys/types.h>  /* for open() */
#include <sys/stat.h>   /* for open() */
#include <fcntl.h>      /* for open() */
#include <unistd.h>     /* for read() */

#if 0
// Typical usage:
//   export SSL_CERT_FILE=/etc/ssl/certs/jsd-cacert.pem
//   lynx -auth=jsd:sesame -source
//       'https://server.example.com/pvt/urandom?bytes=12'  | wc
//
// Any 'http' URL will get rewritten to 'https' by the server.

// NOTES:
//
//  cd /etc/apache2/mods-enabled
//  ln -s ../mods-available/rewrite.load ./
//
// Add to server "default" config
// and also to port-443 "default-ssl" config
//             AllowOverride FileInfo AuthConfig Limit Indexes
//
// Add to "default-ssl" config (and not the other config)
//        <LocationMatch "/pvt/" >
//          Require valid-user
//        </LocationMatch>
//
// Add +ExecCGI here, in both config files:
//        <Directory /var/www/>
//                Options Indexes FollowSymLinks MultiViews +ExecCGI
//
// /etc/init.d/apache2 restart
//
// install -d /var/www/pvt
// chown jsd.jsd /var/www/pvt
// install .htaccess in that directory

// install -d /usr/local/apache/passwd/
// htpasswd -c /usr/local/apache/passwd/passwords jsd
//
#endif

using namespace std;

// parse a QUERY_STRING, looking for bytes=NNNN
// return NNNN as an integer
int parse(string query) {
  string howmany("0");
  string::size_type where;
  while (query.length()) {
    where = query.find('&');
    string item(query);
    if (where != string::npos) {
      item = query.substr(0, where);
      query = query.substr(1+where);
    } else {
      query = "";
    }
    ///cout << "item: " << item << endl;
    string key(item);
    string val;
    where = item.find('=');
    if (where != string::npos){
      key = item.substr(0, where);
      val = item.substr(1+where);
    } else {
      howmany = key;
      key = "";
    }
    if (key == "bytes") howmany = val;
  }
  return atoi(howmany.c_str());
}

map<string,string> char2map(char * const * environ){
  map<string,string> rslt;
  int env_count(0);
  for (char * const * eptr = environ; *eptr; eptr++){
    string item = *eptr;
    env_count++;
    string::size_type where = item.find('=');
    string key(item);
    string val;
    if (where != string::npos) {
      key = item.substr(0, where);
      val = item.substr(1+where);
    }
    rslt[key] = val;
  }
  return rslt;
}

// note that cerr scribbles onto /var/log/apache2/error.log

int main(){
  map<string,string> env = char2map(environ);

// For status code definitions, see
// http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
  if (env["HTTPS"] != "on") {
    cout << "Status: 501 forbidden" << endl;
    cout << "Content-Type: text/plain" << endl;
    cout << endl;
    cout << "HTTPS secure connection is required" << endl;
    exit(0);
  }

// same as REQUEST_URI but with query stripped off:
  string script = env["SCRIPT_NAME"];
  string query = env["QUERY_STRING"];

// Authorization should be configured in .htaccess and handled by
// the server, long before control gets to this point.
// You might think we could attempt a 401 "authorization required"
// but that would be futile, not least because a proper 401 MUST
// contain a challenge suitable for retrying with authorization,
// and we are nowhere near the layer where that stuff happends.
  if (env["AUTH_TYPE"] != "Basic") {
    cout << "Status: 402 not authorized." << endl;
    cout << "Content-Type: text/plain" << endl;
    cout << endl;
    cout << "Possibly bad userID or password." << endl;
    cout << "Possibly bad .htaccess configuration on server." << endl;
    exit(0);
  }

// starting from "/a/b/c/random.cgi", strip it down to just "random"
// strip last "." and everything after that:
  string::size_type where = script.rfind('.');
  if (where != string::npos) script = script.substr(0, where);
// strip last "/" and everything before that:
  where = script.rfind('/');
  if (where != string::npos) script = script.substr(1+where);

// be super-defensive about what files we will open:
  char const* fname;
  if      (script == "random") fname = "/dev/random";
  else if (script == "urandom") fname = "/dev/urandom";
  else {
    cout << "Status: 501 not implemented" << endl;
    cout << "Content-Type: text/plain" << endl;
    cout << endl;
    cout << "Don't know what to do with '"
       << script << "'." << endl;
    exit(0);
  }

// open the input file:
  int ifd = open(fname, O_RDONLY);
  if (ifd < 0) {
    cout << "Status: 503 service unavailable" << endl;
    cout << "Content-Type: text/plain" << endl;
    cout << endl;
    cout << "Cannot open '"
       << fname << "'." << endl;
    exit(0);
  }

// OK, all set up, time to do the work.
  int howmany(0);
  howmany = parse(query);

// Header for data stream:
  cout << "Content-Type: application/octet-stream" << endl;
  cout << "Content-Length: " << howmany << endl;
  cout << endl;         // end of headers

  while (howmany > 0) {
    char buf[1024];
    int todo = howmany;
    if (todo > int(sizeof(buf))) todo = sizeof(buf);
    int got = read(ifd, buf, todo);
    if (got == 0) {
      cerr << "random.cgi: unexpected Eof" << endl;
      break;
    }
    if (got < 0) {
      cerr << "random.cgi: input error" << endl;
      break;
    }
    int wrote = write(1, buf, got);
    if (wrote < 0) {
      cerr << "random.cgi: output error" << endl;
      break;
    }
    if (wrote != got) {
      cerr << "random.cgi: short write: " << wrote
        << "  instead of: " << got
        << endl;
      break;
    }
    howmany -= got;
  }

#if 0
  cout << "Content-Type: text/plain" << endl;
  cout << endl;         // end of headers


  cout << "query:  " << query << endl;
  cout << "script: " << script << endl;
  cout << "howmany: " << howmany << endl;
#endif
}
